# This module contains the functionality for replacing the javaScript-relevant tags inside of tasks.
# hg_js_tag_processor is called by javaScriptUtils.py to replace the tags.
# The javaScript-functions contained may also just be called by any user, as their importation is solely based on
# a call being contained in the processed javaScript-part.
import re
import os
import base64
import sys
from .messages import info, error, abort

# loads picture under given path and returns the dataURL for it without enclosing quotes
def picPathToDataURL(path):
    imageType = os.path.splitext(path)[1][1:]
    with open(path, "rb") as fl:
        return "data:image/" + imageType + ";base64," + str(base64.encodestring(fl.read()))[2:-3]

CHECK_BOX_SEPARATOR = "HGCHECKBOXSEPARATOR"

def escapeLineBreaks(text):
    return text.replace("\n","\\n")

# JS-code for starting an ACE-editor (source-code-editor).
CODE_EDITOR = """ 
if(typeof(HGEditorNum) == "undefined"){HGEditorNum = 0;} // note that this numbering deals with copies when reviewing answers


function HGEditor(container, language){
/**
* Creates an editor inside the given container (replaces inner HTML) with highlighting in the given language.
*/
    container.innerHTML = "<div id=HGInnerEditor" + HGEditorNum + "></div>"; // container used as a handle
    var editor = ace.edit("HGInnerEditor" + HGEditorNum);
    editor.setTheme("ace/theme/monokai");
    editor.session.setMode("ace/mode/" + language);
    editor.setOptions({minLines: 10, maxLines: 10000}); // maxLines have to be set, else always size of about one line
    editor.setFontSize(16);
    container.editor = editor;
    HGEditorNum += 1;
    return container;
}
"""

VARIABLE_TABLE_ROW_SEPARATOR = "HGJSVARIABLETABLEROWSEPARATOR";
VARIABLE_TABLE_COLUMN_SEPARATOR = "HGJSVARIABLETABLECOLUMNSEPARATOR";

VARIABLE_TABLE = """
if(typeof(HGVariableTableNum) == "undefined"){HGVariableTableNum = 0;} // note that this numbering deals with copies when reviewing answers

function HGVariableTable(container, baseX, baseY, defaultEntry){
  var currentNum = HGVariableTableNum;
  container.innerHTML = '""" + """
<style type="text/css">
.HGVTableDatColumn{num}{
  float: left;
  width: 90%;
}

.HGVTableNavColumn{num}{
  float: left;
  width: 10%;
}

.HGVTableRow{num}:after{
  content: "";
  display: table;
  clear: both;
}

</style>
<div id = "HGVTableContainer{num}" class = "HGVTableRow{num}">
<div class = "HGVTableNavColumn{num}">
<p align = "center">
<button type = "button" onclick="HGElement(\\'HGVTableContainer{num}\\').extendY()">Y+</button>
</p>
<p align = "center">
<button type = "button" onclick="HGElement(\\'HGVTableContainer{num}\\').shrinkX()">X-</button> 
<button type = "button" onclick="HGElement(\\'HGVTableContainer{num}\\').extendX()">X+</button>
</p>
<p align = "center">
<button type = "button" onclick="HGElement(\\'HGVTableContainer{num}\\').shrinkY()">Y-</button>
</p>
</div>
<div class = "HGVTableDatColumn{num}">
</div>
</div>
""".replace("\n","\\n") + """'.replace(/{num}/g, currentNum)

  var innerContainer = HGElement("HGVTableContainer" + currentNum)

  var values = [[defaultEntry]] // note we start with a 1x1 table, so the steps which read the DOM do not need to detect a size of 0 in any dimension
  var xSize = 1;
  var ySize = 1;
  var rowSeparator = '""" + VARIABLE_TABLE_ROW_SEPARATOR + """';
  var columnSeparator = '""" + VARIABLE_TABLE_COLUMN_SEPARATOR + """';
  function readEntries(){
    var rows = innerContainer.children[1].children[0].children[0].children; // container (with control buttons and table) -> right part (table) -> body -> rows
    for(var y = 0; y < ySize; y++){
    var cells = rows[y].children;
    for(var x = 0; x < xSize; x++){
      values[x][y] = cells[x].children[0].value; // cell -> input -> value
    }
    }
    values;
  }
  function reDisplay(){
    var tableContent = "<table>";
    for(var y = 0; y < ySize; y++){
    tableContent += "<tr>"
    for(var x = 0; x < xSize; x++){
      var insertValue = values[x][y].replace(/"/g,'&#34;').replace(/'/g,"&#39;") // escape quotation marks so they do not interfere with the DOM attributes in the next step
      tableContent += '<th><input class="ilc_qinput_TextInput" type="text" spellcheck="false" autocomplete="off" autocorrect="off" value="' + insertValue + '" autocapitalize="off" size = 10></th>'
    }
    tableContent += "</tr>"
    }
    tableContent += "</table>";
    innerContainer.children[1].innerHTML = tableContent; // container -> right part (table)
  }
  innerContainer.extendX = function(){
    readEntries();
    if(values.length < xSize + 1){
    values.push([]);
    }
    while(values[xSize].length < ySize){
    values[xSize].push(defaultEntry);
    }
    xSize += 1;
    reDisplay();
  }
  innerContainer.extendY = function(){
    readEntries();
    ySize += 1;
    for(var x = 0; x < xSize; x++){
    while(values[x].length < ySize){
      values[x].push(defaultEntry);
    }
    }
    reDisplay();
  }
  innerContainer.shrinkX = function(){
    readEntries();
    xSize = Math.max(1, xSize - 1);
    reDisplay();
  }
  innerContainer.shrinkY = function(){
    readEntries();
    ySize = Math.max(1, ySize - 1);
    reDisplay();
  }

  reDisplay(); // just to set up the dom for readEntries() during expansion

  for(var x = 0; x < baseX - 1; x++){ // rather inefficient way of getting to initial size, should be changed at some point
    innerContainer.extendX();
  }
  for(var y = 0; y < baseY - 1; y++){
    innerContainer.extendY();
  }

  container.HGGetter = function(){
    readEntries();
    var output = "";
    var firstX = true;
    for(var x = 0; x < xSize; x++){
      if(!firstX){
        output += columnSeparator;
      }
      firstX = false;
      var firstY = true;
      for(var y = 0; y < ySize; y++){
        if(!firstY){
          output += rowSeparator;
        }
        firstY = false;
        output += values[x][y];
      }
    }
    return output;
  }
  container.HGSetter = function(value){
    values = [];
    var columns = value.split(columnSeparator);
    xSize = columns.length;
    for(var x = 0; x < xSize; x++){
    values.push([]);
    var rows = columns[x].split(rowSeparator);
    ySize = rows.length; // should write the same value over and over again, but performance impact should be relatively low (even compared to alternatives)
    for(var y = 0; y < ySize; y++){
      values[x].push(rows[y]);
    }
    }
    reDisplay();
  }
  container.HGGetEntries = function(){
    readEntries();
    var valCopy = [];
    for(var x = 0; x < xSize; x++){
    valCopy.push(values[x].slice(0, ySize));
    }
    return valCopy;
  }
  container.HGSetEntries = function(newVals){
    values = [];
    xSize = newVals.length;
    ySize = newVals[0].length;
    for(var x = 0; x < xSize; x++){
    values.push([...newVals[x]]);
    }
    reDisplay();
  }
  return container;
}
"""

graphNodeTypes = {"simpleCircle" : ("simpleCircle","SimpleCircle.png"), "doubleCircle" : ("doubleCircle","DoubleCircle.png"), "simpleDiamond" : ("simpleDiamond","SimpleDiamond.png"), "doubleDiamond" : ("doubleDiamond","DoubleDiamond.png")}
# JS-code for starting a canvas with graph-editing functionality
GRAPH_EDITOR ="""
if(typeof(HGGraphEditorNum) == "undefined"){HGGraphEditorNum = 0;} // note that this numbering deals with copies when reviewing answers

function HGGraphEditor(container, directedness, usedTypes){
   /* the container is used as a handle for the graph editor */

   container.innerHTML =""" + """'<style type="text/css">

  #graphEditorMenu' + HGGraphEditorNum + ' {
    position: fixed;
    z-index: 9999; /* Most times is 2000 used as middle */
    visibility: hidden;
    opacity: 0;
    
    padding: 0px;
    font-family: sans-serif;
    font-size: 11px;
    background: #fff;
    color: #555;
    border: 1px solid #C6C6C6;
    
    -webkit-transition: opacity .5s ease-in-out;
    -moz-transition: opacity .5s ease-in-out;
    transition: opacity .5s ease-in-out;
    
    -webkit-box-shadow: 2px 2px 2px 0px rgba(143, 144, 145, 1);
    -moz-box-shadow: 2px 2px 2px 0px rgba(143, 144, 145, 1);
    box-shadow: 2px 2px 2px 0px rgba(143, 144, 145, 1);
  }

  #graphEditorMenu' + HGGraphEditorNum + ' a {
    display: block;
    color: #555;
    text-decoration: none;
    padding: 6px 8px 6px 30px;
    width: 250px;
    position: relative;
  }

  #graphEditorMenu' + HGGraphEditorNum + ' a img,
  #graphEditorMenu' + HGGraphEditorNum + ' a i.fa {
    height: 20px;
    font-size: 17px;
    width: 20px;
    position: absolute;
    left: 5px;
    top: 2px;
  }

  #graphEditorMenu' + HGGraphEditorNum + ' a span {
    color: #BCB1B3;
    float: right;
  }

  #graphEditorMenu' + HGGraphEditorNum + ' a:hover {
    color: #fff;
    background: #3879D9;
  }

  #graphEditorMenu' + HGGraphEditorNum + ' hr {
    border: 1px solid #EBEBEB;
    border-bottom: 0;
  }
    
    #graphEditorNetwork' + HGGraphEditorNum + ' {
      width: 600px;
      height: 400px;
      background: #FFFFFF;
      border: 1px solid lightgray;
    }
  </style>
  <div id="graphEditorMenu' + HGGraphEditorNum + '">
      <a href="#n" class="HGGraphControl" onclick="HGElement(\\'graphEditorNetwork' + HGGraphEditorNum + '\\').HGEraseElement()">
        Löschen (Muss zweifach geklickt werden!)
      </a>
      <a href="#n" class="HGGraphControl"  onclick="HGElement(\\'graphEditorNetwork' + HGGraphEditorNum + '\\').HGChangeType()">
  Art ändern
      </a>
  </div>

  <div id="graphEditorNetwork' + HGGraphEditorNum + '"></div>'""".replace("\n","\\n") + """

  var nodes = [];
  var edges = [];

  var edgeStyles = ['to', 'from', 'to,from', '']; // warning, this will get cut down in the lines right below depending on the directedness of the graph
  if(directedness == "directed"){ // arrow to and arrows from
    edgeStyles = edgeStyles.slice(0, 2);
  }
  else if(directedness == "undirected"){ // only ''
    edgeStyles = edgeStyles.slice(3, 4);
  }
  else{ // in all other cases use any
    edgeStyles = edgeStyles.slice(0, 3); // arrows to, from or both
  }

  var nodePicNames = [];
  var nodePics = [];
  var nodePicNamesReverse = {};

  for (pic of usedTypes){
    nodePicNames.push(pic[0]);
    nodePics.push(pic[1]);
  }

  for (var a = 0; a < nodePicNames.length; a++){
    nodePicNamesReverse[nodePicNames[a]] = a; 
  }

  var fontOffset = -40; // assuming 100px x 100px images this offset puts the labels into the centre of the image

  nodes = new vis.DataSet(nodes);
  edges = new vis.DataSet(edges);

  // create a network
  var networkContainer = HGElement('graphEditorNetwork' + HGGraphEditorNum);
  var data = {
    nodes: nodes,
    edges: edges
  };
  var options = {};
  var network = new vis.Network(networkContainer, data, options);
  network.physics.setOptions({enabled : false})

  var erasePrimed = false;
  var lastSelected = undefined;
  var selectedNodes = function(){
    return network.getSelectedNodes().length != 0;
  }
  
  networkContainer.HGEraseElement = function(){
    if(erasePrimed == true){
      network.deleteSelected();
      menuStyle.visibility = "hidden";
      erasePrimed = false;
    }
    else{
      erasePrimed = true;
    }
  }
  
  networkContainer.HGChangeType = function(){
    if(selectedNodes()){
      var sel = network.getSelectedNodes()[0];
      var useStyle = (nodes.get(sel).HGType + 1) % nodePics.length;
      nodes.update([{id : sel, HGType : useStyle, image : nodePics[useStyle]}]);
    }
    else{
      var sel = network.getSelectedEdges()[0];
      var useStyle = (edges.get(sel).HGType + 1) % edgeStyles.length;
      edges.update([{id : sel, HGType : useStyle, arrows : edgeStyles[useStyle]}]);
    }
  }
  var selectAt = function(e, select = true){
    var sel = network.getNodeAt(e.pointer.DOM);
    if(sel != undefined){
      if(select){
        network.selectNodes([sel]);
      }
    }
    else{
      sel = network.getEdgeAt(e.pointer.DOM);
      if(sel != undefined){
        if(select){
          network.selectEdges([sel]);
        }
      }
      else{
        return undefined;
      }
    }
    return sel;
  }
  
  
  //context menu modified from one made by Stephan Stanisic (license free)
  var menuStyle = HGElement('graphEditorMenu' + HGGraphEditorNum).style;
  network.addEventListener('oncontext', function(e) {
    menuStyle.visibility = "hidden";
    
    if(!selectAt(e)){
      return;
    }

    var posX = e.event.clientX;
    var posY = e.event.clientY;
    menu(posX, posY);
    catchEv = e;
  }, false);
  container.addEventListener('contextmenu', function(e) {
    e.preventDefault();
  }, false);
  network.addEventListener('click', function(e) {
    var madeCon = false;
    if(e.event.changedPointers[0].shiftKey){
      var selA = lastSelected;
      var selB = selectAt(e, false);
      if(selA && selB){
        edges.add([{from : selA, to : selB, arrows : "to", HGType : 0, label : ""}])
        network.selectNodes([lastSelected])
        madeCon = true;
      }
    }
    if(!madeCon){
      lastSelected = network.getSelectedNodes()[0];
    }
    menuStyle.opacity = "0";
    erasePrimed = false;
    menuStyle.visibility = "hidden";
  }, false);
  network.addEventListener('doubleClick', function(e) {
    var sel = network.getNodeAt(e.pointer.DOM);
    if(sel != undefined){
      var newLabel = prompt("Beschriftung eintragen:", nodes.get(sel).label) || nodes.get(sel).label;
      nodes.update({id : sel, label : newLabel})
    }
    else{
      sel = network.getEdgeAt(e.pointer.DOM);
      if(sel != undefined){
        var newLabel = prompt("Beschriftung eintragen:", edges.get(sel).label) || edges.get(sel).label;
        edges.update({id : sel, label : newLabel})
      }
      else{
        nodes.add([{id : nodes.get().length + 1 , x : e.pointer.canvas.x, y : e.pointer.canvas.y, label:'Knoten', shape:'image', HGType:0, image: nodePics[0], font : {vadjust : fontOffset}}])
      }
    }
  }, false);

  var menu = function(x, y) {
    menuStyle.top = y + "px";
    menuStyle.left = x + "px";
    menuStyle.visibility = "visible";
    menuStyle.opacity = "1";
  }


  // Note: due to the encoding/decoding used when saving and loading, using utf-16 code points for saving and loading did not work properly the last time I tried (the inputs were different in the results screen)

  var fromStringWithLength = function(inp, pos, lenLen){
    var len = parseInt(inp.slice(pos, pos + lenLen), 36);
    pos += lenLen;
    var out = inp.slice(pos, pos + len);
    return [len + lenLen, out];
  }

  container.HGRefit = function(){
    network.fit();
  }

  var replaceGraphContents = function(newNodes, newEdges){
    nodes.clear();
    edges.clear();
    for(var a = 0; a < newNodes.length; a++){
      nodes.add(newNodes[a]);
    }
    for(var a = 0; a < newEdges.length; a++){
      edges.add(newEdges[a]);
    }
    network.redraw();
    container.HGRefit();
  }

  container.HGSetter = async function(inp){
    var dPos = 0;
    var readPos = 0;
    var nodeNum;
    [dPos, nodeNum] = fromStringWithLength(inp, readPos, 1);
    nodeNum = parseInt(nodeNum);
    readPos += dPos;
    var newNodes = [];
    for(var a = 0; a < nodeNum; a++){
      var useX, useY, useID, useType, useLabel;
      [dPos, useX] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useY] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useID] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useType] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useLabel] = fromStringWithLength(inp, readPos, 2);
      readPos += dPos;
      newNodes.push({x : parseInt(useX), y : parseInt(useY), id : parseInt(useID), HGType : parseInt(useType), label : useLabel, font : {vadjust : fontOffset}, shape : "image", image : nodePics[parseInt(useType)]});
    }

    var edgeNum;
    [dPos, edgeNum] = fromStringWithLength(inp, readPos, 1);
    edgeNum = parseInt(edgeNum);
    readPos += dPos;
    var newEdges = [];
    for(var a = 0; a < edgeNum; a++){
      var useFrom, useTo, useType, useLabel;
      [dPos, useFrom] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useTo] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useType] = fromStringWithLength(inp, readPos, 1);
      readPos += dPos;
      [dPos, useLabel] = fromStringWithLength(inp, readPos, 2);
      readPos += dPos;
      newEdges.push({from : parseInt(useFrom), to : parseInt(useTo), HGType : parseInt(useType), label : useLabel, arrows : edgeStyles[parseInt(useType)]});
    }
    replaceGraphContents(newNodes, newEdges);
  };

  var toStringWithLength = function(inp, lenLen){
    inp = String(inp);
    var inpLen = (inp.length).toString(36);
    inpLen = "0".repeat(lenLen - inpLen.length) + inpLen;
    return inpLen + inp;
  }  

  container.HGGetter = function(){
    var out = "";
    var saveNodes = nodes.get();
    var nodePositions = network.getPositions();
    out += toStringWithLength(saveNodes.length,1);
    for(var a = 0; a < saveNodes.length; a++){
      var curNode = saveNodes[a];
      var curPos = nodePositions[curNode.id]
      out += toStringWithLength(Math.round(curPos.x), 1);
      out += toStringWithLength(Math.round(curPos.y), 1);
      out += toStringWithLength(curNode.id, 1);
      out += toStringWithLength(curNode.HGType, 1);
      out += toStringWithLength(curNode.label, 2);
    }
    var saveEdges = edges.get();
    out += toStringWithLength(saveEdges.length, 1);
    for(var a = 0; a < saveEdges.length; a++){
      var curEdge = saveEdges[a];
      out += toStringWithLength(curEdge.from, 1);
      out += toStringWithLength(curEdge.to, 1);
      out += toStringWithLength(curEdge.HGType, 1);
      out += toStringWithLength(curEdge.label, 2);
    }
    return out;
  };

  container.HGSetGraph = function(inp){
    var newNodes = [];
    var newEdges = [];
    var getNodes = inp.nodes;
    var getEdges = inp.edges;
    for(var a = 0; a < getNodes.length; a++){
      var curNode = getNodes[a];
      var useType = nodePicNamesReverse[curNode.type];
      var newNode = {id : curNode.id + 1, HGType :useType, label : curNode.label, font : {vadjust : fontOffset}, shape : "image", image : nodePics[useType]}
      if(curNode.x != undefined && curNode.y != undefined){
        newNode["x"] = curNode.x;
        newNode["y"] = curNode.y;
      }
      newNodes.push(newNode);
    }
    for(var a = 0; a < getEdges.length; a++){
      var curEdge = getEdges[a];
      var useType = 0;
      if(directedness != "undirected" && curEdge.bidirectional){
        useType = 2;
      }
      newEdges.push({from : curEdge.from + 1, to : curEdge.to + 1, arrows : edgeStyles[useType], HGType : useType, label : curEdge.label})
    }
    replaceGraphContents(newNodes, newEdges);
  }

  container.HGGetGraph = function(){
    var convNodes = [];
    var convEdges = [];
    var edgeList = {forwards : {}, backwards : {}};
    var getNodes = nodes.get();
    var getEdges = edges.get();
    var reID = {}; // nodes get renumbered so the IDs are easier to deal with
    var curID = 0;
    for(var a = 0; a < getNodes.length; a++){
      edgeList["forwards"][a] = [];
      edgeList["backwards"][a] = [];
      var curNode = getNodes[a];
      reID[curNode.id] = curID;
      convNodes.push({'id' : curID, 'label' : curNode.label, 'type' : nodePicNames[curNode.HGType]})
      curID++;
    }
    for(var a = 0; a < getEdges.length; a++){
      var curEdge = getEdges[a];
      var directed = !(curEdge.arrows == 'to,from' || curEdge.arrows == '');
      var from = reID[curEdge.from];
      var to = reID[curEdge.to];
      if(curEdge.arrows == 'from'){
        var tmp = from;
        from = to;
        to = tmp;
      }
      edgeList["forwards"][from].push(a);
      edgeList["backwards"][to].push(a);
      if(!directed){
        edgeList["forwards"][to].push(a);
        edgeList["backwards"][from].push(a);
      }
      convEdges.push({'from' : from, 'to' : to, 'label' : curEdge.label, 'bidirectional' : !directed});
    }
    return {"nodes" : convNodes, "edges" : convEdges, "edgeList" : edgeList}
  }

  HGGraphEditorNum++;
  return container;
}
"""

# JS-code for starting a canvas with drawing functionality
DRAWING_CANVAS = """ 
if(typeof(HGCanvasNum) == "undefined"){HGCanvasNum = 0;} // note that this numbering deals with copies when reviewing answers

function HGDrawingCanvas(container, useWidth, useHeight, maxStack){
/**
* Creates a canvas with drawing functionality inside the given container (replaces inner HTML) along with UI-components.
*/
    container.innerHTML = // container used as as handle
      "<canvas id = 'HGDrawingCanvas" + HGCanvasNum + "' width='" + useWidth + "' height='" + useHeight + "' style = 'border:2px solid;'></canvas>"
      +"<div style = 'float:left; width:250px; padding:10px;'>" // column
        +"<p>Farbe:</p>"
        +"<input id = 'HGDrawingCanvasColorPicker" + HGCanvasNum + "' type = 'color' value = '#009000'>"
        +"<p>Strichstärke:</p>"
        +"<input id = 'HGDrawingCanvasLineSizeRange" + HGCanvasNum + "' type = 'range' min = '1' max = '20' style = 'width:80%;'>"
        +"<input id = 'HGDrawingCanvasUndo" + HGCanvasNum + "' type = 'button' value = 'Rückgängig machen' size = '23' onclick = 'HGElement(\\"HGDrawingCanvas" + HGCanvasNum + "\\").HGUndo()'>"
        +"<input id = 'HGDrawingCanvasRedo" + HGCanvasNum + "' type = 'button' value = 'Wiederherstellen' size = '23' onclick = 'HGElement(\\"HGDrawingCanvas" + HGCanvasNum + "\\").HGRedo()'>"
        +"<input id = 'HGDrawingCanvasClear" + HGCanvasNum + "' type = 'button' value = 'Löschen' size = '23' onclick = 'HGElement(\\"HGDrawingCanvas" + HGCanvasNum + "\\").HGClear()'>"
        +"<p>Werkzeug:</p>"
        +"<input type='radio' name='HGDrawingCanvasToolSelector" + HGCanvasNum + "' value='pencil' checked='checked'>Stift<br>" // note that pencil is selected by default (additional enforcement through checked attribute)
        +"<input type='radio' name='HGDrawingCanvasToolSelector" + HGCanvasNum + "' value='textStamp'>Textstempel<br>"
        +"<input id = 'HGDrawingCanvasTextInput" + HGCanvasNum + "'  class='ilc_qinput_TextInput' type='text' spellcheck='false' autocomplete='off' autocorrect='off' autocapitalize='off' size='25' maxlength='100' value = 'Hier Text zum Stempeln eingeben!'>" // class copied from ILIAS
      +"</div>"
    var picker = HGElement("HGDrawingCanvasColorPicker" + HGCanvasNum);
    var canvas = HGElement("HGDrawingCanvas" + HGCanvasNum);
    var clearer = HGElement("HGClearDrawingCanvas" + HGCanvasNum);
    var context = canvas.getContext("2d");
    var slider = HGElement("HGDrawingCanvasLineSizeRange" + HGCanvasNum);
    var toolSelectors = document.getElementsByName("HGDrawingCanvasToolSelector" + HGCanvasNum);
    var textInput = HGElement("HGDrawingCanvasTextInput" + HGCanvasNum);
    var drawingMode = "pencil";
    textInput.style.display = "none"; // hide input for text stamp in beginning (pencil is selected by default)
    var toolSelectionProcessor = function() { // to be called on change of selected tool
        drawingMode = this.value;
        switch(this.value){
            case "pencil":
                textInput.style.display = "none";
                break;
            case "textStamp":
                textInput.style.display = "";
                break;
            default:
        }
    }
    for (var tool = 0; tool < toolSelectors.length; tool++){
        toolSelectors[tool].addEventListener('change', toolSelectionProcessor);
    }
    canvas.HGUndoPoints = []; // every time something is drawn the picture is pushed onto this list (acts as stack) beforehand
    canvas.HGRedoPoints = []; // on every undo the undone picture is pushed onto this list (acts as stack) and it is emptied whenever something is drawn
    var drawing = false; // indicates whether currently drawing, notice closure

    var pushToUndo = function () {
        canvas.HGUndoPoints.push(canvas.toDataURL());
        if(canvas.HGUndoPoints.length == maxStack + 1){ // note, that when maxStack == -1 this will allow an infinite stack
          canvas.HGUndoPoints.shift();
        }
        canvas.HGRedoPoints = [];
    }

    var mouseProcessor = function (type, event) { // used for processing mouse inputs on the canvas
        var mpx = event.clientX - canvas.getClientRects()[0].x;//mouse positions
        var mpy = event.clientY - canvas.getClientRects()[0].y;
        switch(drawingMode){
            case "pencil":
                if (type == 'move') { // mouse moved
                    if (drawing) { // if currently drawing, draw line
                        prevX = currX;
                        prevY = currY;
                        currX = mpx;
                        currY = mpy;
                        context.strokeStyle = picker.value;
                        context.lineWidth = slider.value;
                        context.lineTo(currX, currY);
                        context.stroke();
                    }
                }
                else if (type == 'down') { // start to draw when mouse pressed
                    currX = mpx;
                    currY = mpy;
                    pushToUndo();
                    drawing = true;
                    context.beginPath();
                    context.lineJoin = "round"; // important to avoid weird spiky lines
                }
                else if (type == 'up' || type == "out") { // moved out of canvas or stopped pressing mouse, stop drawing
                    context.closePath();
                    drawing = false;
                }
                break;
            case "textStamp":
                if (type == 'down') { // start to draw when mouse pressed
                    pushToUndo();
                    context.font = slider.value * 2 + "px Arial"; // Arial assumed to always be available, multiplier to get a reasonable range for text sizes
                    context.fillStyle = picker.value;
                    context.fillText(textInput.value, mpx, mpy); 
                }
                break;
        }
    }
    canvas.addEventListener("mousemove", function (event) {
        mouseProcessor('move', event)
    },
    false);
    canvas.addEventListener("mousedown", function (event) {
        mouseProcessor('down', event)
    },
    false);
    canvas.addEventListener("mouseup", function (event) {
        mouseProcessor('up', event)
    },
    false);
    canvas.addEventListener("mouseout", function (event) {
        mouseProcessor('out', event)
    },
    false);
    container.HGGetter = function() {
      return canvas.toDataURL();
    }
    container.HGSetter = async function(dataURL) {
      var img = new Image;
      img.onload = function(){context.drawImage(img, 0, 0)};
      img.src = dataURL;
    }
    canvas.HGUndo = function(){
      if(canvas.HGUndoPoints.length != 0){
        canvas.HGRedoPoints.push(canvas.toDataURL());
        container.HGSetter(canvas.HGUndoPoints.pop());
      }
      else{
        alert("Kein Schritt kann rückgängig gemacht werden!")
      }
    }
    canvas.HGRedo = function(){
      if(canvas.HGRedoPoints.length != 0){
        canvas.HGUndoPoints.push(canvas.toDataURL());
        container.HGSetter(canvas.HGRedoPoints.pop());
      }
      else{
        alert("Kein Schritt kann wiederhergestellt werden!")
      }
    }
    canvas.HGClear = function(){
      context.rect(0, 0, canvas.width, canvas.height); // doing this only once sometimes results in only partial results (???), twice alway worked, so we do it three times
      context.rect(0, 0, canvas.width, canvas.height);
      context.rect(0, 0, canvas.width, canvas.height);
      context.fillStyle = "white";
      context.fill();
    }
    canvas.HGClear(true); // initially fully white
    HGCanvasNum += 1;
    return container;
}
"""

# JS-code for spinning up a v86-VM
VIRTUAL_MACHINE = """

if(typeof(HGVMNum) == "undefined"){HGVMNum = 0;} // note that this numbering deals with copies when reviewing answers

// A BETTER SOLUTION SHOULD PROBABLY BE FOUND, AS NOW TWO VMS WILL LOAD WHEN REVIEWING ANSWERS (AND BOTH ISOS WILL BE SENT!)


function HGMakeVM(container, biosURL, vgaBiosURL, ISOURL, memory, vMemory){
/**
* Creates a VM inside the given container (replaces inner HTML).
* The bios, vga-bios and iso are all given as URLs (normally data-URLs).
* The memory and vMemory are given in bytes.
*/
    container.innerHTML = '<div id="HGVMScreenContainer' + HGVMNum + '"><div style="white-space: pre; font: 14px monospace; line-height: 14px"></div><canvas style="display: none"></canvas></div>';
    var emulator = new V86Starter({
        memory_size: memory,
        vga_memory_size: vMemory,
        screen_container: document.getElementById("HGVMScreenContainer" + HGVMNum),
        bios: {
            url: biosURL		
        },
        vga_bios: {
            url: vgaBiosURL		
        },
        cdrom: {
            url: ISOURL		
        },
            autostart: true,
        });
    var outputBuffer = ""; // buffer for serial-terminal-output
    var redirection = undefined; // a redirection function, which output from the serial terminal will be sent to
    var loggedIn = false; // initially a login is required for the serial terminal, this shows whether this has happened already
    emulator.add_listener("serial0-output-char", function(char) // note the hardcoded username root, and that '/root% ' is always considered a request for input (sort of bad, but hard to do another way (wait? (how long?)))
    {
        outputBuffer += char;
	if(outputBuffer.endsWith("login:") && !loggedIn){ // automatic login
            container.HGToTerminal("root\\n");
            loggedIn = true;
        }
	else if(outputBuffer.endsWith("/root% ") && !(redirection === undefined)){
            redirection(outputBuffer.slice(0,-7)); // remove the '/root% ' and the last line break
        }
    });
    container.HGFromTerminal = function(){
        return outputBuffer;
    }
    container.HGToTerminal = function(commands){
        outputBuffer = "";
        emulator.serial0_send(commands);
    }
    container.HGRedirectOutput = function (redirector){
        redirection = redirector;
    }
    container.HGGetInnerVM = function(){
        // returns the v86 emulator object
        return emulator;
    }
    return container;
}
"""

def InterleaveWithTags(text, eType, tagIDs, placeholder):
    """Splits the text at the placeholder string and inserts html-elements of the given type with the given tagIDs as 'HGID's. Most of the time span and div are use, depending on whether the element is supposed to be inline or separated. May fail if number of tagIDs and placeholders in the text not the same."""
    taskParts = text.split(placeholder)
    pos = 0
    text = ""
    while(pos < len(tagIDs)):
        text += taskParts[pos]
        text += "<" + eType + " HGID=\"" + str(tagIDs[pos]) + "\"></" + eType + ">"
        pos = pos + 1
    text += taskParts[-1]
    return text

def readID(found, counter, tagType, pos = 0):
    if(found.groups()[pos] and len(found.groups()[pos])>2):
        gotID = found.groups()[pos][1:-1]
        if(not (gotID in counter)):
            counter[gotID] = 0
        counter[gotID] += 1;
        return gotID
    else:
        abort("Missing ID for tag of type " + tagType + "! Conversion aborted.") # this should never happen, the regexes should catch this case
      

# One might want to consider somehow doing all the insertions with some common function, as they all have very similar structure.
# As they only share a very small minority of their lines each, this would probably not even result in a reduction of code size.
def hg_js_tag_processor(task):
    """Replaces tags for different inputs with replacable HTML-Elements and creates code for conversion."""
    converters = ""
    converters += """
    if(typeof HGFeedbackMap == 'undefined'){
      HGFeedbackMap = {};
      HGFeedbackOnlyList = [];
    }
    """
    IDCount = {} # we count the HGIDs to warn the user of duplicates
# feedback-only content
    reg = re.compile('\[feedbackOnly(\([^)]+?\))\]([\w\W\s]*?)(\[\/feedbackOnly\])', re.MULTILINE) 
    # \[feedbackOnly(\([^)]+?\))\]- finds the opening tag with HGID, for example [feedbackOnly(myFeedbackContainer)]
    # ([\w\W\s]*?) - finds any content in the middle of the tags, supposed to be the language for highlighting, not greedy
    # (\[\/feedbackOnly\]) - finds closing [/feedbackOnly]
    IDs = []
    feedbackContents = []
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "feedbackOnly")
        content = point.groups()[1]
        IDs.append(ID)
        feedbackContents.append(content)
        converters+= """
        HGFeedbackOnlyList.push(HGElements({HGID})[HGInstance]);
        """.replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGFEEDBACKONLYINPUTPLACE", task)
    taskParts = task.split("HGFEEDBACKONLYINPUTPLACE") # bespoke insertion logic compared to rest, since we can not do the insertions in javascript and we want these to not display when javascript fails
    pos = 0
    task = ""
    while(pos < len(IDs)):
        task += taskParts[pos]
        task += "<div hidden HGID=\"" + str(IDs[pos]) + "\">" + feedbackContents[pos] +"</div>"
        pos = pos + 1
    task += taskParts[-1]
# source-code-editors
    reg = re.compile('\[editor(\([^)]+?\))\]([\w\W]*?)(\[\/editor(\([\w\W\s]*?\))?\])', re.MULTILINE) 
    # \[editor(\([^)]+?\))\] - finds the opening tag with HGID, for example [editor(myEditor)]
    # ([\w\W]*?) - finds any word in the middle of the tags, supposed to be the language for highlighting
    # (\[\/editor(\([\w\W\s]*?\))?\]) - finds closing [/editor(feedback)] or [/editor] with feedback being any string with possible line breaks
    IDs = []
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "editor")
        feedback = point.groups()[3]
        lang = point.groups()[1]
        IDs.append(ID)
        converters += """
        HGTempEditor = HGElements({HGID})[HGInstance];
        HGTempEditor = HGEditor(HGTempEditor, {EditorLang});
        HGTempEditor.setAttribute("HGInput", {HGID});
        HGTempEditor.HGGetter = (function(useEditor){ var editorCopy=useEditor; var func = function(){return editorCopy.getValue()}; return func;})(HGTempEditor.editor);
        HGTempEditor.HGSetter = (function(useEditor){ var editorCopy=useEditor; var func = function(content){editorCopy.setValue(content)}; return func;})(HGTempEditor.editor);
        """.replace("{HGID}", "\"" + ID + "\"").replace("{EditorLang}", "\"" + lang + "\"") # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            feedback = feedback.replace('"', '\\"') # auto-escape double-quotes
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempEditor.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + escapeLineBreaks(feedback) + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGEDITORINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGEDITORINPUTPLACE")    
# text-fields
    reg = re.compile('\[textField(\([^)]+?\))\](\d*)(,\d*)?(\[\/textField(\(.*?\))?\])', re.MULTILINE)
    #\[textField(\([^)]+?\))\] - finds the initial tag with HGID, for example [textField(myTextField)]
    #(\d*)(,\d*)? - finds any integer in the middle of the tags, supposed to be the maximum input size of the field or two comma separated integers with the second integer now being the visual size
    #(\[\/textField(\(.*?\))?\]) - finds closing [/textField(feedback)] or [/textField] with feedback being any string (but shouldn't contain line breaks)
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "textField")
        feedback = point.groups()[4]
        maxSize = point.groups()[1]
        size = point.groups()[2]
        if(size == None):
            size = maxSize
        else:
            size = size[1:]
        IDs.append(ID)
        converters += """
        HGTempField = HGElements({HGID})[HGInstance];
        HGTempField.innerHTML = '<input class="ilc_qinput_TextInput" type="text" spellcheck="false" autocomplete="off" autocorrect="off" autocapitalize="off" size="{fieldSize}" maxlength="{fieldCap}">'; // copied from ILIAS
        HGTempField.setAttribute("HGInput", {HGID});
        HGTempField.HGGetter = (function(useField){ var fieldCopy=useField; var func = function(){return fieldCopy.value}; return func;})(HGTempField.childNodes[0]);
        HGTempField.HGSetter = (function(useField){ var fieldCopy=useField; var func = function(content){fieldCopy.value = content}; return func;})(HGTempField.childNodes[0]);
        """.replace("{HGID}", "\"" + ID + "\"").replace("{fieldSize}", size).replace("{fieldCap}", maxSize) # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            feedback = feedback.replace('"', '\\"') # auto-escape double-quotes
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempField.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + feedback + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGTEXTFIELDINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGTEXTFIELDINPUTPLACE") 
# inline-text-fields
    reg = re.compile('\[inlineTextField(\([^)]+?\))\](\d*)(,\d*)?(\[\/inlineTextField(\(.*?\))?\])', re.MULTILINE)
    #\[inlineTextField(\([^)]+?\))\] - finds the initial tag with HGID, for example [inlineTextField(myTextField)]
    #(\d*)(,\d*)? - finds any integer in the middle of the tags, supposed to be the maximum input size of the field or two comma separated integers with the second integer now being the visual size
    #(\[\/inlineTextField(\(.*?\))?\]) - finds closing [/inlineTextField(feedback)] or [/inlineTextField] with feedback being any string without line breaks
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "inlineTextField")
        feedback = point.groups()[4]
        maxSize = point.groups()[1]
        size = point.groups()[2]
        if(size == None):
            size = maxSize
        else:
            size = size[1:]
        IDs.append(ID)
        converters += """
        HGTempField = HGElements({HGID})[HGInstance];
        HGTempField.innerHTML = '<input class="ilc_qinput_TextInput" type="text" spellcheck="false" autocomplete="off" autocorrect="off" autocapitalize="off" size="{fieldSize}" maxlength="{fieldCap}">'; // copied from ILIAS
        HGTempField.setAttribute("HGInput", {HGID});
        HGTempField.HGGetter = (function(useField){ var fieldCopy=useField; var func = function(){return fieldCopy.value}; return func;})(HGTempField.childNodes[0]);
        HGTempField.HGSetter = (function(useField){ var fieldCopy=useField; var func = function(content){fieldCopy.value = content}; return func;})(HGTempField.childNodes[0]);
        """.replace("{HGID}", "\"" + ID + "\"").replace("{fieldSize}", size).replace("{fieldCap}", maxSize) # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            feedback = feedback.replace('"', '\\"') # auto-escape double-quotes
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempField.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + feedback + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGINLINETEXTFIELDINPUTPLACE", task)
    task = InterleaveWithTags(task, "span", IDs, "HGINLINETEXTFIELDINPUTPLACE") 
# text-areas
    reg = re.compile('\[textArea(\([^)]+?\))\]((\d)*)(\[\/textArea(\([\w\W\s]*?\))?\])', re.MULTILINE)
    # \[textArea(\([^)]+?\))\] - finds the initial tag with HGID, for example [textArea(myTextArea)]
    # ((\d)*) - finds any integer in the middle of the tags, supposed to be the maximum input length (omittable)
    # (\[\/textArea(\([\w\W\s]*?\))?\]) - finds closing [/textArea(feedback)] or [/textArea] with feedback being any string possibly with line-breaks
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "textArea")
        feedback = point.groups()[4]
        maxSize = point.groups()[1]
        IDs.append(ID)
        maxSizePart = ""; # will be inserted into attributes, standard is no attribute added
        if(maxSize != ""): # in case maxSize was given, add attribute
            maxSizePart = 'maxlength = ' + maxSize;
        converters += """
        HGTempArea = HGElements({HGID})[HGInstance];
        HGTempArea.innerHTML = '<textarea cols="80" rows="6" {maxSizePart}></textarea>'; // not from ILIAS, tinyMCE used there (but that requires API-key etc.)
        HGTempArea.setAttribute("HGInput", {HGID});
        HGTempArea.HGGetter = (function(useArea){ var areaCopy=useArea; var func = function(){return areaCopy.value}; return func;})(HGTempArea.childNodes[0]);
        HGTempArea.HGSetter = (function(useArea){ var areaCopy=useArea; var func = function(content){areaCopy.value = content}; return func;})(HGTempArea.childNodes[0]);
        """.replace("{HGID}", "\"" + ID + "\"").replace("{maxSizePart}", maxSizePart) # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            feedback = feedback.replace('"', '\\"') # auto-escape double-quotes
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempArea.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + escapeLineBreaks(feedback) + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGTEXTAREAINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGTEXTAREAINPUTPLACE") 
# checkboxes
    reg = re.compile('\[(\\*?)checkBoxes(\([^)]+?\))\]([\w\W]*?)(\[\/checkBoxes(\([^\)]*?\))?\])', re.MULTILINE) # unsure whether to use checkboxes or checkBoxes
    #\[(\\*?)checkBoxes(\([^)]+?\))\] - finds the initial tag with HGID, for example [checkBoxes(myCheckBoxes)] including the possible alias-star
    #([\w\W]*?) - takes everything the middle of the tags, supposed to be the comma-separated selectable values
    #(\[\/checkBoxes(\([^\)]*?\))?\]) - finds closing [/checkBoxes(feedback)] or [/checkBoxes] with feedback being the comma-separated correct answers (written the same as between the tags)
    extraReg = re.compile('([^,]*)([\W]*)', re.MULTILINE) # for the contained comma-separated values
    #([^,]*) - everything up to a comma (should be the checkable options) 
    #([\W]*) - any non-alphanumeric characters afterwards (filters out the commata and spaces etc.)
    IDs = []
    once = True # if true, first checkBox (in this part of the task, copy when reviewing answers will have a new first checkBox)
    for point in re.finditer(reg, task):
        if(once): # if first checkbox add HGCheckBoxSeparator as a global var (could be done multiple times too)
            once = False
            converters += "HGCheckBoxSeparator = '" + CHECK_BOX_SEPARATOR + "';"
        options = []
        values = []
        for option in re.finditer(extraReg, point.groups()[2]): # read the comma-separated options and put them in a list
            options.append(option.groups()[0].replace("\\","\\\\"))
            values.append(option.groups()[0].replace("\\","\\\\"))
        options = options[0:-1] # the loop before adds an empty option in the end
        values = values[0:-1]
        if(point.groups()[0] != None): # alias star contained: change options and values
            aliasReg = re.compile("([\w\W]*):([\w\W]*)", re.MULTILINE)
            for optionNum in range(len(options)):
                for split in re.finditer(aliasReg, options[optionNum]):
                    options[optionNum] = split.groups()[0]
                    values[optionNum] = split.groups()[1]
        ID =  readID(point, IDCount, "radioButtons", pos = 1)
        feedback = point.groups()[4]
        IDs.append(ID)
        inner = ""
        for optionNum in range(len(options)):
            inner += "<p><input type='checkbox' name='" + ID +"checkbox' value='" + values[optionNum] + "'>" + options[optionNum] + "</p>"
        converters += """
        HGTempCheckBoxes = HGElements({HGID})[HGInstance];
        HGTempCheckBoxes.innerHTML = "{inner}"
        HGTempCheckBoxes.setAttribute("HGInput", {HGID}); //note that we will use names for identification to avoid the forced lowercasing from the attributes
        HGTempCheckBoxes.HGGetter = (function(useBoxes){ 
        var boxes=useBoxes.children; 
          var func = function(){
            var output = "";
            for (var boxNum = 0; boxNum < boxes.length; boxNum++){ // go through all checkboxes, concatenate content of checked boxes with the given separator inbetween
              if(boxes[boxNum].childNodes[0].checked){
                output += boxes[boxNum].childNodes[0].value + '{separator}';
              }
            }
            if(output != ""){
              output = output.slice(0, -'{separator}'.length);
            }
            return output;
          };
        return func;})(HGTempCheckBoxes);
        HGTempCheckBoxes.HGSetter = (function(useBoxes){ 
          var boxes=useBoxes.children; 
          var func = function(content){
            var inputs = content.split('{separator}');
            for (var boxNum = 0; boxNum < boxes.length; boxNum++){ // go through the boxes and check everyone contained in the list of the answers after splitting with the given separator
              if(inputs.includes(boxes[boxNum].childNodes[0].value)){
                boxes[boxNum].childNodes[0].checked = true;
              }
            }
          };
          return func;
        })(HGTempCheckBoxes);
        """.replace("{inner}", inner).replace("{HGID}", "\"" + ID + "\"").replace("{separator}", CHECK_BOX_SEPARATOR) # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempCheckBoxes.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + feedback.replace(",", CHECK_BOX_SEPARATOR) + "\"").replace("{HGID}", "\"" + ID + "\"") # also replace the commata inside the feedback with the separator
    task = re.sub(reg, "HGCHECKBOXINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGCHECKBOXINPUTPLACE") 
# radio buttons
    reg = re.compile('\[(\\*?)radioButtons(\([^)]+?\))\]([\w\W]*?)(\[\/radioButtons(\([^\)]*?\))?\])', re.MULTILINE)
    #\[(\\*?)radioButtons(\([^)]+?\))\] - finds the initial tag with HGID, for example [radioButtons(myRadioButtons)] including the possible alias-star
    #([\w\W]*?) - takes everything the middle of the tags, supposed to be the comma-separated selectable values
    #(\[\/radioButtons(\([^\)]*?\))?\]) - finds closing [/radioButtons(feedback)] or [/radioButtons] with feedback being the correct answer (written the same as between the tags)
    extraReg = re.compile('([^,]*)([\W]*)', re.MULTILINE) # for the contained comma-separated values
    #([^,]*) - everything up to a comma (should be the checkable options) 
    #([\W]*) - any non-alphanumeric characters afterwards (filters out the commata and spaces etc.)
    IDs = [];
    for point in re.finditer(reg, task):
        options = []
        values = []
        for option in re.finditer(extraReg, point.groups()[2]): # read the comma-separated options and put them in a list
            options.append(option.groups()[0].replace("\\","\\\\"))
            values.append(option.groups()[0].replace("\\","\\\\"))
        options = options[0:-1] # the loop before adds an empty option in the end
        values = values[0:-1]
        if(point.groups()[0] != None): # alias star contained: change options and values
            aliasReg = re.compile("([\w\W]*):([\w\W]*)", re.MULTILINE)
            for optionNum in range(len(options)):
                for split in re.finditer(aliasReg, options[optionNum]):
                    options[optionNum] = split.groups()[0]
                    values[optionNum] = split.groups()[1]
        ID =  readID(point, IDCount, "radioButtons", pos = 1)
        feedback = point.groups()[4]
        IDs.append(ID)
        inner = ""
        for optionNum in range(len(options)):
            inner += "<p><input type='radio' name='" + ID +"radio' value='" + values[optionNum] + "'>" + options[optionNum] + "</p>"
        converters += """
        HGTempRadios = HGElements({HGID})[HGInstance];
        HGTempRadios.innerHTML = "{inner}"
        HGTempRadios.setAttribute("HGInput", {HGID}); //note that we will use names for identification to avoid the forced lowercasing from the attributes
        HGTempRadios.HGGetter = (function(useRadios){ var radios=useRadios.children; 
        var func = function(){
                for (var radioNum = 0; radioNum < radios.length; radioNum++){ // go through all associated radio buttons and return the content of the first checked one
                    if(radios[radioNum].childNodes[0].checked){
                        return radios[radioNum].childNodes[0].value;
                    }
                }
                return "";
            };
        return func;})(HGTempRadios);
        HGTempRadios.HGSetter = (function(useRadios){
          var radios=useRadios.children; 
            var func = function(content){
              for (var radioNum = 0; radioNum < radios.length; radioNum++){ // go through all radio buttons and check the first one which corresponds to the given content
                  if(radios[radioNum].childNodes[0].value == content){
                      radios[radioNum].childNodes[0].checked = true;
                      return;
                  }
              }
            };
          return func;
        })(HGTempRadios);
        """.replace("{inner}", inner).replace("{HGID}", "\"" + ID + "\"") # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempRadios.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + feedback + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGRADIOSINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGRADIOSINPUTPLACE") 
# selections
    reg = re.compile('\[(\\*?)selection(\([^)]+?\))\]([\w\W]*?)(\[\/selection(\([^\)]*?\))?\])', re.MULTILINE)
    #\[(\\*?)selection(\([^)]+?\))\] - finds the initial tag with HGID, for example [selection(mySelection)] including the possible alias-star
    #([\w\W]*?) - takes everything the middle of the tags, supposed to be the comma-separated selectable values
    #(\[\/selection(\([^\)]*?\))?\]) - finds closing [/selection(feedback)] or [/selection] with feedback being the correct answer (written the same as between the tags)
    extraReg = re.compile('([^,]*)([\W]*)', re.MULTILINE) # for the contained comma-separated values
    #([^,]*) - everything up to a comma (should be the checkable options) 
    #([\W]*) - any non-alphanumeric characters afterwards (filters out the commata and spaces etc.)
    IDs = [];
    for point in re.finditer(reg, task):
        options = []
        values = []
        for option in re.finditer(extraReg, point.groups()[2]): # read the comma-separated options and put them in a list
            options.append(option.groups()[0].replace("\\","\\\\"))
            values.append(option.groups()[0].replace("\\","\\\\"))
        options = options[0:-1] # the loop before adds an empty option in the end
        values = values[0:-1]
        if(point.groups()[0] != None): # alias star contained: change options and values
            aliasReg = re.compile("([\w\W]*):([\w\W]*)", re.MULTILINE)
            for optionNum in range(len(options)):
                for split in re.finditer(aliasReg, options[optionNum]):
                    options[optionNum] = split.groups()[0]
                    values[optionNum] = split.groups()[1]
        ID = readID(point, IDCount, "selection", pos = 1)
        feedback = point.groups()[4]
        IDs.append(ID)
        inner = "<select name='" + ID +"radio'>"
        for optionNum in range(len(options)):
            inner += "<option value='" + values[optionNum] + "'>" + options[optionNum] + "</option>"
        inner += "</select>"
        converters += """
        HGTempSelection = HGElements({HGID})[HGInstance];
        HGTempSelection.innerHTML = "{inner}"
        HGTempSelection.setAttribute("HGInput", {HGID}); //note that we will use names for identification to avoid the forced lowercasing from the attributes
        HGTempSelection.HGGetter = (function(useSelection){ var selection = useSelection.children[0];
            var func = function(){
                return selection.value;
            }
            return func;
        })(HGTempSelection);
        HGTempSelection.HGSetter = (function(useSelection){ var selection = useSelection.children[0];
            var func = function(val){
                selection.value = val;
            }
            return func;
        })(HGTempSelection);
        """.replace("{inner}", inner).replace("{HGID}", "\"" + ID + "\"") # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempSelection.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + feedback + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGSELECTIONINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGSELECTIONINPUTPLACE") 
# drawing surfaces (canvas with mouse-drawing)
    reg = re.compile('\[drawingCanvas(\([^)]+?\))\]((\d*),(\d*))?(,)?(\d*)?(\[\/drawingCanvas(\(.*?\))?\])', re.MULTILINE)
    #\[drawingCanvas(\([^)]+?\))\] - finds the initial tag with HGID, for example [drawingCanvas(myDrawingCanvas)]
    #((\d*),(\d*))?(;)?(\d*) - finds the width, height and maximum stack depth
        #((\d*),(\d*))? finds the width and height - only together
        #(,)? this allows the width and height to be separated from the stack depth with a comma, but it is not required if width and height are not there
        #(\d*)? finds the maximum stack depth
        # note that this only works, because width and height can only be given together, so for any number of arguments from 0 to 3 there is only one possibility for the selected parameters to include
    #(\[\/drawingCanvas(\(.*?\))?\]) - finds closing [/drawingCanvas(feedback)] or [/drawingCanvas] with feedback being a (local) path to a correct picture
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "drawingCanvas")
        width = point.groups()[2]
        height = point.groups()[3]
        stackDepth = point.groups()[5]
        feedbackPath = point.groups()[7]
        if(width == None): # therefore height == None
            width = "400" # default 400x400
            height = "400"
        if(stackDepth == None):
            stackDepth = "-1" # default infinite stack
        IDs.append(ID)
        converters += """
        HGTempCanvas = HGElements({HGID})[HGInstance];
        HGTempCanvas = HGDrawingCanvas(HGTempCanvas, {width}, {height}, {stackDepth});
        HGTempCanvas.setAttribute("HGInput", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{width}", width).replace("{height}", height).replace("{stackDepth}", stackDepth) # using 'format' bad idea due to curly braces
        if feedbackPath != None: # feedback is optional, so check is needed
            feedbackPath = feedbackPath[1:-1]
            imageType = os.path.splitext(feedbackPath)[1][1:]
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempCanvas.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{feedback}", "\"" + picPathToDataURL(feedbackPath) + "\"").replace("{HGID}", "\"" + ID + "\"")
    task = re.sub(reg, "HGDRAWINGCANVASINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGDRAWINGCANVASINPUTPLACE") 
# graph editors
    reg = re.compile('\[graphEditor(\([^)]+?\))\]([\w\W]*?)(\[\/graphEditor\])', re.MULTILINE)
    #\[graphEditor(\([^)]+?\))\] - finds the initial tag with HGID, for example [graphEditor(myGraphEditor)]
    #([\w\W]*?) - grabs all the options inside, separated later
    #(\[\/graphEditor\]) - finds closing [/graphEditor]
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "graphEditor")
        IDs.append(ID)
        args = point.groups()[1]
        args = re.sub("\s", "", args).split(",")
        directedness = args[0]
        args = args[1:]
        basep = os.path.dirname(os.path.abspath(__file__));
        pics = "[";
        for arg in args:
            if(arg in graphNodeTypes):
                nType, nPath = graphNodeTypes[arg];
                pics += "[\"" + nType + "\",\"" + picPathToDataURL(basep + "/GraphNodePics/" + nPath) + "\"],"
            else:
                info("Unknown type of graph node requested:" + arg);
        pics = pics[0:-1] + "]"; 
        converters += """
        HGTempGraphEditor = HGElements({HGID})[HGInstance];
        HGTempGraphEditor = HGGraphEditor(HGTempGraphEditor, {directedness} , {pics});
        HGTempGraphEditor.setAttribute("HGInput", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{pics}", pics).replace("{directedness}", "\"" + directedness + "\"") # using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGGRAPHEDITORINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGGRAPHEDITORINPUTPLACE")   
# variable Tables
    reg = re.compile('\[variableTable(\([^)]+?\))\]([\d]+),([\d]+)(,(.+?))?(\[\/variableTable(\([^\)]*?\))?\])', re.MULTILINE)
    #\[variableTable(\([^)]+?\))\] - finds the initial tag with HGID, for example [variableTable(myVTable)]
    #([\d]+),([\d]+)(,(.+?))? - grabs two numbers, the initial dimensions, and optionally whatever comes after another comma before the ending tag, the default entry
    #(\[\/variableTable(\([^\)]*?\))?\]) - finds closing [/variableTable(feedback)] with feedback being the rows separated by semicolons and the cells being separated by colons
    IDs = [];
    once = True # if true, first checkBox (in this part of the task, copy when reviewing answers will have a new first checkBox)
    for point in re.finditer(reg, task):
        if(once): # if first checkbox add HGCheckBoxSeparator as a global var (could be done multiple times too)
            once = False
            converters += "HGVTableRowSeparator = '" + VARIABLE_TABLE_ROW_SEPARATOR + "';"
            converters += "HGVTableColumnSeparator = '" + VARIABLE_TABLE_COLUMN_SEPARATOR + "';"
        ID = readID(point, IDCount, "variableTable")
        IDs.append(ID)
        defX = point.groups()[1]
        defY = point.groups()[2]
        defEntry = point.groups()[4];
        feedback = point.groups()[6];
        if(defEntry == None):
            defEntry = "-"
        defEntry = defEntry.replace('"', "&#34").replace("'", "&#39") # escape quotation marks in html style to prevent HGVariableTable() from breaking
        converters += """
        HGTempVariableTable = HGElements({HGID})[HGInstance];
        HGTempVariableTable = HGVariableTable(HGTempVariableTable, {defX} , {defY}, {defEntry});
        HGTempVariableTable.setAttribute("HGInput", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{defX}", defX).replace("{defY}", defY).replace("{defEntry}", "\"" + defEntry + "\"") # using 'format' bad idea due to curly braces
        if feedback != None: # feedback is optional, so check is needed (or replace feedback by empty string beforehand)
            feedback = feedback[1:-1] # remove enclosing brackets
            feedback = feedback.replace(";", VARIABLE_TABLE_COLUMN_SEPARATOR).replace(",", VARIABLE_TABLE_ROW_SEPARATOR).replace('"', "&#34").replace("'", "&#39")
            converters += """
            HGFeedbackMap[{HGID}] = {feedback};
            if(HGInstance == HGFeedbackInstance){
              HGTempVariableTable.HGSetter(HGFeedbackMap[{HGID}]);
            }
            """.replace("{HGID}", "\"" + ID + "\"").replace("{feedback}", "\"" + feedback + "\"") # also replace the commata inside the feedback with the separator
    task = re.sub(reg, "HGVARIABLETABLEINPUTPLACE", task)
    task = InterleaveWithTags(task, "p", IDs, "HGVARIABLETABLEINPUTPLACE")   
# buttons
    reg = re.compile('\[button(\([^)]+?\))\]([\w\W]+?),([^,]*?)(\[\/button\])', re.MULTILINE) # note that the inscription of buttons can not contain commas...
    #\[button(\([^)]+?\))\] - finds the initial tag with HGID, for example [button(myButton)]
    #([\w\W]+?) - takes everything the middle of the tags, except for everything after the last comma (see next part), supposed to be the javaScript to execute on button-press
    #([^,]*?) - takes everything in the middle of the tags after the last comma, supposed to be the title of the button
    #(\[\/button\]) - finds closing [/button]
    IDs = [];
    once = True # if true, first button (in this part of the task, copy when reviewing answers will have a new first button)
    for point in re.finditer(reg, task):
        if(once): # if first button add initialization of button function list and access(could be done multiple times too)
            once = False
            converters += """
            if(typeof HGButtonFunctionList == 'undefined'){
              HGButtonFunctionList=[[],[]];
              HGGlobalize(HGButtonFunctionList, "HGButtonFunctionList");
              HGExecuteButtonFunc = function(number, instance){HGButtonFunctionList[instance][number]()};
              HGGlobalize(HGExecuteButtonFunc, "HGExecuteButtonFunc");
            } """
        ID = readID(point, IDCount, "button")
        toExecute = point.groups()[1]
        #toExecute = toExecute.replace('"',"&quot;").replace("'","&apos;") # automatic html escaping for both types of quotation marks so they can be used here # not needed anymore
        title = point.groups()[2]
        IDs.append(ID)
        converters += """
        HGCurrentButtonFuncCloser = function(){return function(){{toExecute}}};
        HGButtonFunctionList[HGInstance].push(HGCurrentButtonFuncCloser());
        HGTempButton = HGElements({HGID})[HGInstance];
        HGTempButton.innerHTML = '<button type="button" onclick="HGExecuteButtonFunc(' + (HGButtonFunctionList[HGInstance].length - 1) + ',' + HGInstance + ')">{title}</button>'; // assign function from last button of current instance
        HGTempButton.setAttribute("HGUtility", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{toExecute}", toExecute).replace("{title}", title) # using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGBUTTONINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGBUTTONINPUTPLACE")  
# feedback-buttons
    reg = re.compile('\[feedbackButton(\([^)]+?\))\](\[\/feedbackButton\])', re.MULTILINE) # note that the inscription of buttons can not contain commas...
    #\[feedbackButton(\([^)]+?\))\] - finds the initial tag with HGID, for example [feedbackButton(myFButton)]
    #(\[\/feedbackButton\]) - finds closing [/feedbackButton]
    IDs = [];
    once = True # if true, first button (in this part of the task, copy when reviewing answers will have a new first button)
    for point in re.finditer(reg, task):
        if(once): # if first feedback-button add globalized function for revealing all feedback and blocking the answer
            once = False
            converters += """
            if(typeof HGFeedbackBlocker == 'undefined'){
              var blockEval = HGBlockingEvaluator;
              var HGFButtonFunc = function(){
                HGRegisterEvaluator(blockEval); // block further input
                HGMetaData["viewedSolution"]="yes"; // marker to block further input on upcoming viewings of the task
                HGSaveAll(); // immediately save the changed metadata and send it to the server to minimize abusability
                HGSaveToServer(); 
                var inputs = document.querySelectorAll("[HGInput]") // get all inputs
                for (var InpNum = 0; InpNum < inputs.length; InpNum++){
                  if(typeof HGFeedbackMap[inputs[InpNum].getAttribute("HGInput")] != "undefined"){ // if feedback exists, set content to feedback
                    inputs[InpNum].HGSetter(HGFeedbackMap[inputs[InpNum].getAttribute("HGInput")] );
                  }
                }
                for (var FBONum = 0; FBONum < HGFeedbackOnlyList.length; FBONum++){ // go through feedback-only-parts and display them
                  HGFeedbackOnlyList[FBONum].removeAttribute("hidden");
                }
              };
              HGGlobalize(HGFButtonFunc, "HGFeedbackBlocker");
            } """
        ID = readID(point, IDCount, "feedbackButton")
        IDs.append(ID)
        converters += """
        HGTempFButton = HGElements({HGID})[HGInstance];
        HGTempFButton.innerHTML = '<button type="button" onclick="if(confirm(\\'Hierdurch verzichten Sie auf eine Korrektur aller Eingaben, welche Sie in dieser Aufgabe tätigen!\\')){HGFeedbackBlocker()}">Musterlösung anzeigen</button>';
        HGTempFButton.setAttribute("HGUtility", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"")# using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGFBUTTONINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGFBUTTONINPUTPLACE")  
# text-display (wrapping for pre-tag)
    reg = re.compile('\[text(\([^)]+?\))\]([\w\W]*?)(\[\/text\])', re.MULTILINE)
    #\[text(\([^)]+?\))\] - finds the initial tag with HGID, for example [text(myText)]
    #([\w\W]*?) - takes everything the middle of the tags, supposed to be the standard content of the text-display
    #(\[\/text\]) - finds closing [/text]
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "text")
        standardContent = point.groups()[1]
        IDs.append(ID)
        converters += """
        HGTempText = HGElements({HGID})[HGInstance];
        HGTempText.innerHTML = '<pre>{standardContent}</pre>';
        HGTempText.setAttribute("HGUtility", {HGID});
        HGTempText.HGGetter = (function(useText){ var textCopy=useText; var func = function(){return textCopy.textContent}; return func;})(HGTempText.childNodes[0]);
        HGTempText.HGSetter = (function(useText){ var textCopy=useText; var func = function(content){textCopy.textContent = content}; return func;})(HGTempText.childNodes[0]);
        """.replace("{HGID}", "\"" + ID + "\"").replace("{standardContent}", standardContent) # using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGTEXTINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGTEXTINPUTPLACE")  
# images
    reg = re.compile('\[image(\([^)]+?\))\]([\w\W]*?)(\[\/image\])', re.MULTILINE)
    #\[image(\([^)]+?\))\] - finds the initial tag with HGID, for example [image(myImage)]
    #([\w\W]*?) - takes everything the middle of the tags, supposed to be the (relative) path to the image
    #(\[\/image\]) - finds closing [/image]
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "image")
        picPath = point.groups()[1]
        IDs.append(ID)
        converters += """
        HGTempImage = HGElements({HGID})[HGInstance];
        HGTempImage.innerHTML = '<img src={dataURL}>';
        HGTempImage.setAttribute("HGUtility", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{dataURL}", "\"" +  picPathToDataURL(picPath) + "\"") # using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGTEXTINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGTEXTINPUTPLACE")  
# VMs
    reg = re.compile('\[virtualMachine(\([^)]+?\))\]([\w\W]+?),([\w\W]+?),([\w\W]+?),([\w\W]+?),([\w\W]+?)(\[\/virtualMachine\])', re.MULTILINE)
    #\[virtualMachine(\([^)]+?\))\] - finds the initial tag with HGID, for example [virtualMachine(myVirtualMachine)]
    #([\w\W]+?), ... - to find the fixed amount of comma-separated values:
    # 1. bios
    # 2. vgabios
    # 3. iso
    # 4. memory
    # 5. vmemory
    #(\[\/virtualMachine\]) - finds closing [/virtualMachine]
    IDs = [];
    for point in re.finditer(reg, task):
        ID = readID(point, IDCount, "virtualMachine")
        data = list(point.groups()[1:6])
        IDs.append(ID)
        for fNum in range(3):
            with open(data[fNum].strip(),"rb") as emu_file:
                data[fNum] = str(base64.b64encode(emu_file.read()))[2:-1]
        converters += """
        HGTempVM = HGElements({HGID})[HGInstance];
        HGTempVM = HGMakeVM(HGTempVM, {bios}, {vgaBios}, {iso}, {memory}, {vmemory});
        HGTempVM.setAttribute("HGUtility", {HGID});
        """.replace("{HGID}", "\"" + ID + "\"").replace("{memory}", data[3]).replace("{vmemory}", data[4]).replace("{bios}", "'data:;base64," + data[0] + "'").replace("{vgaBios}", "'data:;base64," + data[1] + "'").replace("{iso}", "'data:;base64," + data[2] + "'")# using 'format' bad idea due to curly braces
    task = re.sub(reg, "HGVIRTUALMACHINEINPUTPLACE", task)
    task = InterleaveWithTags(task, "div", IDs, "HGVIRTUALMACHINEINPUTPLACE")




# error detection
    for(key, val) in IDCount.items():
        if(val>1):
            error("Duplicate ID " + key + " detected. Output will probably have defects.")

    reg = re.compile('\[[^\[]+?\([\w\W]+?\)\]', re.MULTILINE)
    for point in re.finditer(reg, task):
        info("Unprocessed tag-like structure detected (check for correct use, spelling and end-tag):" + point.group())
  
    return (task, converters)
